from scipy.constants import epsilon_0
import numpy as np


def _get_transv_field_gauss_round(sigma, Delta_x, Delta_y, x, y, mathlib):
    exp = mathlib.exp
    sqrt = mathlib.sqrt
    pi = mathlib.pi

    r2 = (x-Delta_x)*(x-Delta_x)+(y-Delta_y)*(y-Delta_y)
    if (r2 < 1e-20):
        temp = sqrt(r2)/(2.*pi*epsilon_0*sigma)  # linearised
    else:
        temp = (1-exp(-0.5*r2/(sigma*sigma)))/(2.*pi*epsilon_0*r2)

    Ex = temp * (x-Delta_x)
    Ey = temp * (y-Delta_y)

    return Ex, Ey


get_transv_field_gauss_round = np.vectorize(_get_transv_field_gauss_round, excluded = ['mathlib'])


def _get_transv_field_gauss_ellip(sigmax, sigmay, Delta_x, Delta_y, x, y, mathlib):

    abs = mathlib.abs
    exp = mathlib.exp
    wfun = mathlib.wfun
    sqrt = mathlib.sqrt
    pi = mathlib.pi
    sqrt_pi = sqrt(pi)

    # I always go to the first quadrant and then apply the signs a posteriori
    # numerically more stable (see http://inspirehep.net/record/316705/files/slac-pub-5582.pdf)

    abx = abs(x - Delta_x)
    aby = abs(y - Delta_y)

    if (sigmax > sigmay):

        S = sqrt(2.*(sigmax*sigmax-sigmay*sigmay))
        factBE = 1./(2.*epsilon_0*sqrt_pi*S)

        etaBE_re = sigmay/sigmax*abx
        etaBE_im = sigmax/sigmay*aby

        zetaBE_re = abx
        zetaBE_im = aby

        w_zetaBE_re, w_zetaBE_im = wfun(zetaBE_re/S, zetaBE_im/S)
        w_etaBE_re, w_etaBE_im = wfun(etaBE_re/S, etaBE_im/S)

        expBE = exp(-abx*abx/(2*sigmax*sigmax)-aby*aby/(2*sigmay*sigmay))

        Ex = factBE*(w_zetaBE_im - w_etaBE_im*expBE)
        Ey = factBE*(w_zetaBE_re - w_etaBE_re*expBE)

    elif (sigmay > sigmax):

        S = sqrt(2.*(sigmay*sigmay-sigmax*sigmax))
        factBE = 1./(2.*epsilon_0*sqrt_pi*S)

        etaBE_re = sigmax/sigmay*aby
        etaBE_im = sigmay/sigmax*abx

        zetaBE_re = aby
        zetaBE_im = abx

        w_zetaBE_re, w_zetaBE_im = wfun(zetaBE_re/S, zetaBE_im/S)
        w_etaBE_re, w_etaBE_im = wfun(etaBE_re/S, etaBE_im/S)

        expBE = exp(-aby*aby/(2*sigmay*sigmay)-abx*abx/(2*sigmax*sigmax))

        Ey = factBE*(w_zetaBE_im - w_etaBE_im*expBE)
        Ex = factBE*(w_zetaBE_re - w_etaBE_re*expBE)
    else:
        Ex = 1./0.
        Ey = 1./0.

    if (x - Delta_x) < 0:
        Ex = -Ex
    if (y - Delta_y) < 0:
        Ey = -Ey

    return Ex, Ey

get_transv_field_gauss_ellip = np.vectorize(_get_transv_field_gauss_ellip, excluded = ['mathlib'])


def _get_Ex_Ey_Gx_Gy_gauss(x, y, sigma_x, sigma_y, min_sigma_diff, skip_Gs, mathlib):

    abs = mathlib.abs
    pi = mathlib.pi
    exp = mathlib.exp

    if (abs(sigma_x-sigma_y) < min_sigma_diff):

        sigma = 0.5*(sigma_x+sigma_y)
        Delta_x = 0.
        Delta_y = 0.

        Ex, Ey = get_transv_field_gauss_round(
            sigma, Delta_x, Delta_y, x, y, mathlib)

        if not skip_Gs:
            Gx = 1/(2.*(x*x+y*y))*(y*Ey-x*Ex+1./(2*pi*epsilon_0*sigma*sigma)
                                   * x*x*exp(-(x*x+y*y)/(2.*sigma*sigma)))
            Gy = 1./(2*(x*x+y*y))*(x*Ex-y*Ey+1./(2*pi*epsilon_0*sigma*sigma)
                                   * y*y*exp(-(x*x+y*y)/(2.*sigma*sigma)))
    else:

        sigma_x = sigma_x
        sigma_y = sigma_y
        Delta_x = 0.
        Delta_y = 0.

        Ex, Ey = get_transv_field_gauss_ellip(
            sigma_x, sigma_y, Delta_x, Delta_y, x, y, mathlib)

        Sig_11 = sigma_x*sigma_x
        Sig_33 = sigma_y*sigma_y
        if not skip_Gs:
            Gx = -1./(2*(Sig_11-Sig_33))*(x*Ex+y*Ey+1./(2*pi*epsilon_0) *
                                          (sigma_y/sigma_x*exp(-x*x/(2*Sig_11)-y*y/(2*Sig_33))-1.))
            Gy = 1./(2*(Sig_11-Sig_33))*(x*Ex+y*Ey+1./(2*pi*epsilon_0) *
                                         (sigma_x/sigma_y*exp(-x*x/(2*Sig_11)-y*y/(2*Sig_33))-1.))

    if skip_Gs:
        return Ex, Ey
    else:
        return Ex, Ey, Gx, Gy


get_Ex_Ey_Gx_Gy_gauss = np.vectorize(_get_Ex_Ey_Gx_Gy_gauss, excluded = ['mathlib'])



import numpy as np

from . import slicing
from . import boost


class ParBoost(object):
    # it is practically a struct
    def __init__(self, phi, alpha):
        self.sphi = np.sin(phi)
        self.cphi = np.cos(phi)
        self.tphi = np.tan(phi)
        self.salpha = np.sin(alpha)
        self.calpha = np.cos(alpha)

    def tobuffer(self):
        buf = [
            self.sphi,
            self.cphi,
            self.tphi,
            self.salpha,
            self.calpha]
        return np.array(buf, dtype=np.float64)


class Sigmas(object):
    def __init__(self, Sig_11_0, Sig_12_0, Sig_13_0,
                 Sig_14_0, Sig_22_0, Sig_23_0, Sig_24_0,
                 Sig_33_0, Sig_34_0, Sig_44_0):

        self.Sig_11_0 = Sig_11_0
        self.Sig_12_0 = Sig_12_0
        self.Sig_13_0 = Sig_13_0
        self.Sig_14_0 = Sig_14_0
        self.Sig_22_0 = Sig_22_0
        self.Sig_23_0 = Sig_23_0
        self.Sig_24_0 = Sig_24_0
        self.Sig_33_0 = Sig_33_0
        self.Sig_34_0 = Sig_34_0
        self.Sig_44_0 = Sig_44_0

    def tobuffer(self):
        buf = [
            self.Sig_11_0,
            self.Sig_12_0,
            self.Sig_13_0,
            self.Sig_14_0,
            self.Sig_22_0,
            self.Sig_23_0,
            self.Sig_24_0,
            self.Sig_33_0,
            self.Sig_34_0,
            self.Sig_44_0]
        return np.array(buf, dtype=np.float64)


def boost_sigmas(Sigma_0, cphi):
    Sigma_0_boosted = Sigmas(
        Sigma_0.Sig_11_0,
        Sigma_0.Sig_12_0/cphi,
        Sigma_0.Sig_13_0,
        Sigma_0.Sig_14_0/cphi,
        Sigma_0.Sig_22_0/cphi/cphi,
        Sigma_0.Sig_23_0/cphi,
        Sigma_0.Sig_24_0/cphi/cphi,
        Sigma_0.Sig_33_0,
        Sigma_0.Sig_34_0/cphi,
        Sigma_0.Sig_44_0/cphi/cphi)
    return Sigma_0_boosted


def int_to_float64arr(val):
    temp = np.zeros(1, (np.float64, {'i64': ('i8', 0)}))
    temp['i64'][0] = val
    return temp


class BB6D_Data(object):
    def __init__(self, q_part,
                 parboost, Sigmas_0_star, N_slices, N_part_per_slice,
                 x_slices_star, y_slices_star, sigma_slices_star,
                 min_sigma_diff, threshold_singular,
                 delta_x, delta_y,
                 x_CO, px_CO, y_CO, py_CO, sigma_CO, delta_CO,
                 Dx_sub, Dpx_sub, Dy_sub, Dpy_sub, Dsigma_sub, Ddelta_sub,
                 enabled):

        self.q_part = q_part
        self.parboost = parboost
        self.Sigmas_0_star = Sigmas_0_star
        self.min_sigma_diff = min_sigma_diff
        self.threshold_singular = threshold_singular
        self.N_slices = N_slices
        self.N_part_per_slice = N_part_per_slice
        self.x_slices_star = x_slices_star
        self.y_slices_star = y_slices_star
        self.sigma_slices_star = sigma_slices_star

        self.delta_x = delta_x
        self.delta_y = delta_y
        self.x_CO = x_CO
        self.px_CO = px_CO
        self.y_CO = y_CO
        self.py_CO = py_CO
        self.sigma_CO = sigma_CO
        self.delta_CO = delta_CO
        self.Dx_sub = Dx_sub
        self.Dpx_sub = Dpx_sub
        self.Dy_sub = Dy_sub
        self.Dpy_sub = Dpy_sub
        self.Dsigma_sub = Dsigma_sub
        self.Ddelta_sub = Ddelta_sub

        self.enabled = enabled

    def tobuffer(self):

        buffer_list = []
        # Buffers corresponding to BB6D struct
        buffer_list.append(np.array([self.q_part], dtype=np.float64))
        buffer_list.append(self.parboost.tobuffer())
        buffer_list.append(self.Sigmas_0_star.tobuffer())
        buffer_list.append(np.array([self.min_sigma_diff], dtype=np.float64))
        buffer_list.append(
            np.array([self.threshold_singular], dtype=np.float64))
        buffer_list.append(int_to_float64arr(self.N_slices))

        buffer_list.append(np.array([self.delta_x], dtype=np.float64))
        buffer_list.append(np.array([self.delta_y], dtype=np.float64))
        buffer_list.append(np.array([self.x_CO], dtype=np.float64))
        buffer_list.append(np.array([self.px_CO], dtype=np.float64))
        buffer_list.append(np.array([self.y_CO], dtype=np.float64))
        buffer_list.append(np.array([self.py_CO], dtype=np.float64))
        buffer_list.append(np.array([self.sigma_CO], dtype=np.float64))
        buffer_list.append(np.array([self.delta_CO], dtype=np.float64))
        buffer_list.append(np.array([self.Dx_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dpx_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dy_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dpy_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dsigma_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Ddelta_sub], dtype=np.float64))
        buffer_list.append(int_to_float64arr(
            {True: 1, False: 0}[self.enabled]))

        buffer_list.append(int_to_float64arr(3))  # offset to N_part_per_slice
        # offset to x_slices_star
        buffer_list.append(int_to_float64arr(2+self.N_slices))
        # offset to y_slices_star
        buffer_list.append(int_to_float64arr(1+2*self.N_slices))
        # offset to sigma_slices_star
        buffer_list.append(int_to_float64arr(0+3*self.N_slices))

        # Buffers corresponding to arrays
        buffer_list.append(np.array(self.N_part_per_slice, dtype=np.float64))
        buffer_list.append(np.array(self.x_slices_star, dtype=np.float64))
        buffer_list.append(np.array(self.y_slices_star, dtype=np.float64))
        buffer_list.append(np.array(self.sigma_slices_star, dtype=np.float64))

        buf = np.concatenate(buffer_list)

        return buf


def BB6D_init(q_part, N_part_tot, sigmaz, N_slices, min_sigma_diff, threshold_singular,
              phi, alpha,
              Sig_11_0, Sig_12_0, Sig_13_0,
              Sig_14_0, Sig_22_0, Sig_23_0,
              Sig_24_0, Sig_33_0, Sig_34_0, Sig_44_0,
              delta_x, delta_y,
              x_CO, px_CO, y_CO, py_CO, sigma_CO, delta_CO,
              Dx_sub, Dpx_sub, Dy_sub, Dpy_sub, Dsigma_sub, Ddelta_sub,
              enabled):

    # Prepare data for Lorentz transformation
    parboost = ParBoost(phi=phi, alpha=alpha)

    # Prepare data with strong beam shape
    Sigmas_0 = Sigmas(Sig_11_0, Sig_12_0, Sig_13_0,
                      Sig_14_0, Sig_22_0, Sig_23_0,
                      Sig_24_0, Sig_33_0, Sig_34_0, Sig_44_0)

    # Boost strong beam shape
    Sigmas_0_star = boost_sigmas(Sigmas_0, parboost.cphi)

    # Generate info about slices
    z_centroids, _, N_part_per_slice = slicing.constant_charge_slicing_gaussian(
        N_part_tot, sigmaz, N_slices)

    # Sort according to z, head at the first position in the arrays
    ind_sorted = np.argsort(z_centroids)[::-1]
    z_centroids = np.take(z_centroids, ind_sorted)
    N_part_per_slice = np.take(N_part_per_slice, ind_sorted)

    # By boosting the strong z and all zeros, I get the transverse coordinates of the strong beam in the ref system of the weak
    boost_vect = np.vectorize(boost.boost, excluded='parboost')
    x_slices_star, px_slices_star, y_slices_star, py_slices_star, sigma_slices_star, delta_slices_star = boost_vect(x=0*z_centroids, px=0*z_centroids,
                                                                                                                    y=0*z_centroids, py=0*z_centroids, sigma=z_centroids, delta=0*z_centroids, parboost=parboost)

    bb6d_data = BB6D_Data(q_part, parboost, Sigmas_0_star, N_slices,
                          N_part_per_slice, x_slices_star, y_slices_star, sigma_slices_star, min_sigma_diff, threshold_singular,
                          delta_x, delta_y, x_CO, px_CO, y_CO, py_CO, sigma_CO, delta_CO,
                          Dx_sub, Dpx_sub, Dy_sub, Dpy_sub, Dsigma_sub, Ddelta_sub,
                          enabled)

    return bb6d_data



import numpy as np
from scipy.special import erf, erfinv


def constant_charge_slicing_gaussian(N_part_tot, sigmaz, N_slices):

    if N_slices > 1:
        # working with intensity 1. and rescling at the end
        Qi = (np.arange(N_slices)/float(N_slices))[1:]

        z_cuts = np.sqrt(2)*sigmaz*erfinv(2*Qi-1.)

        #~ import pdb; pdb.set_trace()

        z_centroids = []
        first_centroid = -sigmaz / \
            np.sqrt(2*np.pi)*np.exp(-z_cuts[0] **
                                    2/(2*sigmaz*sigmaz))*float(N_slices)
        z_centroids.append(first_centroid)
        for ii in range(N_slices-2):
            this_centroid = -sigmaz/np.sqrt(2*np.pi)*(np.exp(-z_cuts[ii+1]**2/(2*sigmaz*sigmaz)) -
                                                      np.exp(-z_cuts[ii]**2/(2*sigmaz*sigmaz)))*float(N_slices)
            # the multiplication times n slices comes from the fact that we have to divide by the slice charge, i.e. 1./N
            z_centroids.append(this_centroid)

        last_centroid = sigmaz / \
            np.sqrt(2*np.pi)*np.exp(-z_cuts[-1] **
                                    2/(2*sigmaz*sigmaz))*float(N_slices)
        z_centroids.append(last_centroid)

        z_centroids = np.array(z_centroids)

        N_part_per_slice = z_centroids*0.+N_part_tot/float(N_slices)
    elif N_slices == 1:
        z_centroids = np.array([0.])
        z_cuts = []
        N_part_per_slice = np.array([N_part_tot])

    else:
        raise ValueError('Invalid number of slices')

    return z_centroids, z_cuts, N_part_per_slice



import numpy as np


def mysign(u): return 2*np.float_(u >= 0)-1.


class Sigmas(object):
    def __init__(self, Sig_11_0, Sig_12_0, Sig_13_0,
                 Sig_14_0, Sig_22_0, Sig_23_0, Sig_24_0,
                 Sig_33_0, Sig_34_0, Sig_44_0):

        self.Sig_11_0 = Sig_11_0
        self.Sig_12_0 = Sig_12_0
        self.Sig_13_0 = Sig_13_0
        self.Sig_14_0 = Sig_14_0
        self.Sig_22_0 = Sig_22_0
        self.Sig_23_0 = Sig_23_0
        self.Sig_24_0 = Sig_24_0
        self.Sig_33_0 = Sig_33_0
        self.Sig_34_0 = Sig_34_0
        self.Sig_44_0 = Sig_44_0

    def tobuffer(self):
        buf = [
            self.Sig_11_0,
            self.Sig_12_0,
            self.Sig_13_0,
            self.Sig_14_0,
            self.Sig_22_0,
            self.Sig_23_0,
            self.Sig_24_0,
            self.Sig_33_0,
            self.Sig_34_0,
            self.Sig_44_0]
        return np.array(buf, dtype=np.float64)


def boost_sigmas(Sigma_0, cphi):
    Sigma_0_boosted = Sigmas(
        Sigma_0.Sig_11_0,
        Sigma_0.Sig_12_0/cphi,
        Sigma_0.Sig_13_0,
        Sigma_0.Sig_14_0/cphi,
        Sigma_0.Sig_22_0/cphi/cphi,
        Sigma_0.Sig_23_0/cphi,
        Sigma_0.Sig_24_0/cphi/cphi,
        Sigma_0.Sig_33_0,
        Sigma_0.Sig_34_0/cphi,
        Sigma_0.Sig_44_0/cphi/cphi)
    return Sigma_0_boosted


def _propagate_Sigma_matrix(Sigmas_at_0, S, threshold_singular=1e-16, handle_singularities=True):

    Sig_11_0 = Sigmas_at_0.Sig_11_0
    Sig_12_0 = Sigmas_at_0.Sig_12_0
    Sig_13_0 = Sigmas_at_0.Sig_13_0
    Sig_14_0 = Sigmas_at_0.Sig_14_0
    Sig_22_0 = Sigmas_at_0.Sig_22_0
    Sig_23_0 = Sigmas_at_0.Sig_23_0
    Sig_24_0 = Sigmas_at_0.Sig_24_0
    Sig_33_0 = Sigmas_at_0.Sig_33_0
    Sig_34_0 = Sigmas_at_0.Sig_34_0
    Sig_44_0 = Sigmas_at_0.Sig_44_0

    #~ Sig_11 = Sig_11_0 + 2.*Sig_12_0*S+Sig_22_0*S*S
    #~ Sig_33 = Sig_33_0 + 2.*Sig_34_0*S+Sig_44_0*S*S
    #~ Sig_13 = Sig_13_0 + (Sig_14_0+Sig_23_0)*S+Sig_24_0*S*S

    Sig_11, Sig_12, Sig_13,\
        Sig_14, Sig_22, Sig_23, Sig_24,\
        Sig_33, Sig_34, Sig_44 = propagate_full_Sigma_matrix_in_drift(
            Sig_11_0, Sig_12_0, Sig_13_0,
            Sig_14_0, Sig_22_0, Sig_23_0, Sig_24_0,
            Sig_33_0, Sig_34_0, Sig_44_0, S)

    R = Sig_11-Sig_33
    W = Sig_11+Sig_33
    T = R*R+4*Sig_13*Sig_13

    # evaluate derivatives
    dS_R = 2.*(Sig_12_0-Sig_34_0)+2*S*(Sig_22_0-Sig_44_0)
    dS_W = 2.*(Sig_12_0+Sig_34_0)+2*S*(Sig_22_0+Sig_44_0)
    dS_Sig_13 = Sig_14_0 + Sig_23_0 + 2*Sig_24_0*S
    dS_T = 2*R*dS_R+8.*Sig_13*dS_Sig_13

    signR = mysign(R)

    if T < threshold_singular and handle_singularities:
        a = Sig_12-Sig_34
        b = Sig_22-Sig_44
        c = Sig_14+Sig_23
        d = Sig_24

        sqrt_a2_c2 = np.sqrt(a*a+c*c)

        if sqrt_a2_c2*sqrt_a2_c2*sqrt_a2_c2 < threshold_singular:
            # equivalent to: if np.abs(c)<threshold_singular and np.abs(a)<threshold_singular:

            if np.abs(d) > threshold_singular:
                cos2theta = np.abs(b)/np.sqrt(b*b+4*d*d)
            else:
                cos2theta = 1.  # Decoupled beam

            costheta = np.sqrt(0.5*(1.+cos2theta))
            sintheta = mysign(b)*mysign(d)*np.sqrt(0.5*(1.-cos2theta))

            dS_costheta = 0.
            dS_sintheta = 0.

            Sig_11_hat = 0.5*W
            Sig_33_hat = 0.5*W

            dS_Sig_11_hat = 0.5*dS_W
            dS_Sig_33_hat = 0.5*dS_W

        else:
            sqrt_a2_c2 = np.sqrt(a*a+c*c)
            cos2theta = np.abs(2*a)/(2*sqrt_a2_c2)
            costheta = np.sqrt(0.5*(1.+cos2theta))
            sintheta = mysign(a)*mysign(c)*np.sqrt(0.5*(1.-cos2theta))

            dS_cos2theta = mysign(
                a)*(0.5*b/sqrt_a2_c2-a*(a*b+2*c*d)/(2*sqrt_a2_c2*sqrt_a2_c2*sqrt_a2_c2))

            dS_costheta = 1/(4*costheta)*dS_cos2theta
            if np.abs(sintheta) > threshold_singular:
                # equivalent to: if np.abs(c)>threshold_singular:
                dS_sintheta = -1/(4*sintheta)*dS_cos2theta
            else:
                dS_sintheta = d/(2*a)

            Sig_11_hat = 0.5*W
            Sig_33_hat = 0.5*W

            dS_Sig_11_hat = 0.5*dS_W + mysign(a)*sqrt_a2_c2
            dS_Sig_33_hat = 0.5*dS_W - mysign(a)*sqrt_a2_c2

    else:

        sqrtT = np.sqrt(T)
        cos2theta = signR*R/sqrtT
        costheta = np.sqrt(0.5*(1.+cos2theta))
        sintheta = signR*mysign(Sig_13)*np.sqrt(0.5*(1.-cos2theta))

        # in sixtrack this line seems to be different different
        #~ sintheta = -mysign((Sig_11-Sig_33))*np.sqrt(0.5*(1.-cos2theta))

        Sig_11_hat = 0.5*(W+signR*sqrtT)
        Sig_33_hat = 0.5*(W-signR*sqrtT)

        dS_cos2theta = signR*(dS_R/sqrtT - R/(2*sqrtT*sqrtT*sqrtT)*dS_T)
        dS_costheta = 1/(4*costheta)*dS_cos2theta

        if np.abs(sintheta) < threshold_singular and handle_singularities:
            # equivalent to to np.abs(Sig_13)<threshold_singular
            dS_sintheta = (Sig_14+Sig_23)/R
        else:
            dS_sintheta = -1/(4*sintheta)*dS_cos2theta

        dS_Sig_11_hat = 0.5*(dS_W + signR*0.5/sqrtT*dS_T)
        dS_Sig_33_hat = 0.5*(dS_W - signR*0.5/sqrtT*dS_T)

    # This will not be exposed in C
    '''
    extra_data = {}
    extra_data['Sig_11'] = Sig_11
    extra_data['Sig_13'] = Sig_13
    extra_data['Sig_33'] = Sig_33
    extra_data['cos2theta'] = cos2theta
    extra_data['T'] = T
    extra_data['R'] = R
    '''
    extra_data = -1.

    #~ if dS_sintheta>1e10:
    #~ import pdb; pdb.set_trace()

    return Sig_11_hat, Sig_33_hat, costheta, sintheta,\
        dS_Sig_11_hat, dS_Sig_33_hat, dS_costheta, dS_sintheta,\
        extra_data

        
propagate_Sigma_matrix = np.vectorize(_propagate_Sigma_matrix)

# def propagate_Sigma_matrix(Sigmas_at_0,
#                                       S, threshold_singular=1e-16, handle_singularities=True):
#
#     Sig_11_hat, Sig_33_hat, costheta, sintheta,\
#         dS_Sig_11_hat, dS_Sig_33_hat, dS_costheta, dS_sintheta,\
#         extra_data_list = np.vectorize(_propagate_Sigma_matrix,
#                                        excluded=['Sigmas_at_0', 'threshold_singular', 'handle_singularities'])(
#             Sigmas_at_0, S, threshold_singular, handle_singularities)
#
#     extra_data = {}
#     for kk in extra_data_list[0].keys():
#         extra_data[kk] = []
#         for ele in extra_data_list:
#             extra_data[kk].append(ele[kk])
#
#     return Sig_11_hat, Sig_33_hat, costheta, sintheta,\
#         dS_Sig_11_hat, dS_Sig_33_hat, dS_costheta, dS_sintheta,\
#         extra_data
#

def propagate_full_Sigma_matrix_in_drift(Sig_11_0, Sig_12_0, Sig_13_0,
                                         Sig_14_0, Sig_22_0, Sig_23_0, Sig_24_0,
                                         Sig_33_0, Sig_34_0, Sig_44_0, S):

    # Can be found in matrix form in A. Wolsky, "Beam dynamics in high energy particle accelerators"

    Sig_11 = Sig_11_0 + 2.*Sig_12_0*S+Sig_22_0*S*S
    Sig_33 = Sig_33_0 + 2.*Sig_34_0*S+Sig_44_0*S*S
    Sig_13 = Sig_13_0 + (Sig_14_0+Sig_23_0)*S+Sig_24_0*S*S

    Sig_12 = Sig_12_0 + Sig_22_0*S
    Sig_14 = Sig_14_0 + Sig_24_0*S
    Sig_22 = Sig_22_0 + 0.*S
    Sig_23 = Sig_23_0 + Sig_24_0*S
    Sig_24 = Sig_24_0 + 0.*S
    Sig_34 = Sig_34_0 + Sig_44_0*S
    Sig_44 = Sig_44_0 + 0.*S

    return Sig_11, Sig_12, Sig_13,\
        Sig_14, Sig_22, Sig_23, Sig_24,\
        Sig_33, Sig_34, Sig_44



import numpy as np
import pysixtrack

import scipy.optimize as so

from pysixtrack.particles import Particles


def get_elems_of_type(line, elemname):

    pyst_ele_type_list = [ele[1] for ele in line]
    indlist = np.where([ss == elemname for ss in pyst_ele_type_list])[0]
    ele_list = [line[ind][2] for ind in indlist]
    name_list = [line[ind][0] for ind in indlist]

    return indlist, name_list, ele_list


class Ring(pysixtrack.Line):
    def __init__(self, line, p0c=1e9):
        self.p0c = p0c
        super().__init__(elements=[elem for label, elem_type, elem in line])

    def one_turn_map(self, coord):

        pcl = Particles(p0c=self.p0c)

        pcl.x = coord[0]
        pcl.px = coord[1]
        pcl.y = coord[2]
        pcl.py = coord[3]
        pcl.zeta = coord[4]
        pcl.delta = coord[5]

        self.track(pcl)

        coord_out = np.array(
            [pcl.x, pcl.px, pcl.y, pcl.py, pcl.sigma, pcl.delta])

        return coord_out

    def _CO_error(self, coord):
        return np.sum((self.one_turn_map(coord)-coord)**2)

    def find_closed_orbit(self, guess=[0., 0., 0., 0., 0., 0.], method='Nelder-Mead'):

        if method == 'get_guess':
            res = type('', (), {})()
            res.x = guess
        else:
            res = so.minimize(self._CO_error, np.array(
                guess), tol=1e-20, method=method)

        pcl = Particles(p0c=self.p0c)

        pcl.x = res.x[0]
        pcl.px = res.x[1]
        pcl.y = res.x[2]
        pcl.py = res.x[3]
        pcl.zeta = res.x[4]
        pcl.delta = res.x[5]

        CO = self.track_elem_by_elem(pcl)

        return CO



from collections import namedtuple
from copy import deepcopy

import numpy as np

# attaching faddeeva to np
from scipy.special import wofz


def wfun(z_re, z_im):
    w = wofz(z_re+1j*z_im)
    return w.real, w.imag


np.wfun = wfun


def count_not_none(*lst):
    return len(lst)-sum(p is None for p in lst)


class Particles(object):
    clight = 299792458
    pi = 3.141592653589793238
    echarge = 1.602176565e-19
    emass = 0.510998928e6
    pmass = 938.272046e6
    epsilon0 = 8.854187817e-12
    mu0 = 4e-7*pi
    eradius = echarge**2/(4*pi*epsilon0*emass*clight**2)
    pradius = echarge**2/(4*pi*epsilon0*pmass*clight**2)
    anumber = 6.02214129e23
    kboltz = 1.3806488e-23

    def _g1(self, mass0, p0c, energy0):
        beta0 = p0c/energy0
        gamma0 = energy0/mass0
        return mass0, beta0, gamma0, p0c, energy0

    def _g2(self, mass0, beta0, gamma0):
        energy0 = mass0*gamma0
        p0c = energy0*beta0
        return mass0, beta0, gamma0, p0c, energy0

    def _f1(self, mass0, p0c):
        sqrt = self._m.sqrt
        energy0 = sqrt(p0c**2+mass0**2)
        return self._g1(mass0, p0c, energy0)

    def _f2(self, mass0, energy0):
        sqrt = self._m.sqrt
        p0c = sqrt(energy0**2-mass0**2)
        return self._g1(mass0, p0c, energy0)

    def _f3(self, mass0, beta0):
        sqrt = self._m.sqrt
        gamma0 = 1/sqrt(1-beta0**2)
        return self._g2(mass0, beta0, gamma0)

    def _f4(self, mass0, gamma0):
        sqrt = self._m.sqrt
        beta0 = sqrt(1-1/gamma0**2)
        return self._g2(mass0, beta0, gamma0)

    def copy(self):
        p = Particles()
        for k, v in list(self.__dict__.items()):
            if type(v) in [np.ndarray, dict]:
                v = v.copy()
            p.__dict__[k] = v
        return p

    def __init__ref(self, p0c, energy0, gamma0, beta0):
        not_none = count_not_none(beta0, gamma0, p0c, energy0)
        if not_none == 0:
            p0c = 1e9
            not_none = 1
            #raise ValueError("Particles defined without energy reference")
        if not_none == 1:
            if p0c is not None:
                self.p0c = p0c
            elif energy0 is not None:
                self.energy0 = energy0
            elif gamma0 is not None:
                self.gamma0 = gamma0
            elif beta0 is not None:
                self.beta0 = beta0
        else:
            raise ValueError(f"""\
            Particles defined with multiple energy references:
            p0c    = {p0c},
            energy0     = {energy0},
            gamma0 = {gamma0},
            beta0  = {beta0}""")

    def __init__delta(self, delta, ptau, psigma):
        not_none = count_not_none(delta, ptau, psigma)
        if not_none == 0:
            self.delta = 0.
        elif not_none == 1:
            if delta is not None:
                self.delta = delta
            elif ptau is not None:
                self.ptau = ptau
            elif psigma is not None:
                self.psigma = psigma
        else:
            raise ValueError(f"""
            Particles defined with multiple energy deviations:
            delta  = {delta},
            ptau     = {ptau},
            psigma = {psigma}""")

    def __init__zeta(self, zeta, tau, sigma):
        not_none = count_not_none(zeta, tau, sigma)
        if not_none == 0:
            self.zeta = 0.
        elif not_none == 1:
            if zeta is not None:
                self.zeta = zeta
            elif tau is not None:
                self.tau = tau
            elif sigma is not None:
                self.sigma = sigma
        else:
            raise ValueError(f"""\
            Particles defined with multiple time deviations:
            zeta  = {zeta},
            tau   = {tau},
            sigma = {sigma}""")

    def __init__chi(self, mratio, qratio, chi):
        not_none = count_not_none(mratio, qratio, chi)
        if not_none == 0:
            self._chi = 1.
            self._mratio = 1.
            self._qratio = 1.
        elif not_none == 1:
            raise ValueError(f"""\
            Particles defined with insufficient mass/charge information:
            chi    = {chi},
            mratio = {mratio},
            qratio = {qratio}""")
        elif not_none == 2:
            if chi is None:
                self._mratio = mratio
                self.qratio = qratio
            elif mratio is None:
                self._chi = chi
                self.qratio = qratio
            elif qratio is None:
                self._chi = chi
                self.mratio = mratio
        else:
            raise ValueError(f"""
            Particles defined with multiple mass/charge information:
            chi    = {chi},
            mratio = {mratio},
            qratio = {qratio}""")

    def __init__(self,
                 s=0., x=0., px=0., y=0., py=0.,
                 delta=None, ptau=None, psigma=None, rvv=None,
                 zeta=None, tau=None, sigma=None,
                 mass0=pmass, q0=1.,
                 p0c=None, energy0=None, gamma0=None, beta0=None,
                 chi=None, mratio=None, qratio=None,
                 partid=None, turn=None, state=None, elemid=None,
                 mathlib=np, **args):

        self._m = mathlib
        self.s = s
        self.x = x
        self.px = px
        self.y = y
        self.py = py
        self.zeta = zeta
        self._mass0 = mass0
        self.q0 = q0
        self._update_coordinates = False
        self.__init__ref(p0c, energy0, gamma0, beta0)
        self.__init__delta(delta, ptau, psigma)
        self.__init__zeta(zeta, tau, sigma)
        self.__init__chi(chi, mratio, qratio)
        self._update_coordinates = True
        self.partid = partid
        self.turn = turn
        self.state = state

    Px = property(lambda p: p.px*p.p0c*p.mratio)
    Py = property(lambda p: p.py*p.p0c*p.mratio)
    Energy = property(lambda p: (p.ptau*p.p0c+p.energy0)*p.mratio)
    Pc = property(lambda p: (p.delta*p.p0c+p.p0c)*p.mratio)
    mass = property(lambda p:  p.mass0*p.mratio)
    beta = property(lambda p:  (1+p.delta)/(1/p.beta0+p.ptau))
    # rvv = property(lambda self: self.beta/self.beta0)
    # rpp = property(lambda self: 1/(1+self.delta))

    rvv = property(lambda self:  self._rvv)
    rpp = property(lambda self:  self._rpp)

    def add_to_energy(self, energy):
        sqrt = self._m.sqrt
        oldrvv = self._rvv
        deltabeta0 = self.delta*self.beta0
        ptaubeta0 = sqrt(deltabeta0**2+2*deltabeta0*self.beta0 + 1)-1
        ptaubeta0 += energy/self.energy0
        ptau = ptaubeta0/self.beta0
        self._delta = sqrt(ptau**2+2*ptau/self.beta0+1)-1
        self._rvv = (1+self.delta)/(1+ptaubeta0)
        self._rpp = 1/(1+self.delta)
        self.zeta *= self._rvv/oldrvv

    delta = property(lambda self: self._delta)

    @delta.setter
    def delta(self, delta):
        sqrt = self._m.sqrt
        self._delta = delta
        deltabeta0 = delta*self.beta0
        ptaubeta0 = sqrt(deltabeta0**2+2*deltabeta0*self.beta0 + 1)-1
        self._rvv = (1+self.delta)/(1+ptaubeta0)
        self._rpp = 1/(1+self.delta)

    psigma = property(lambda self: self.ptau/self.beta0)

    @psigma.setter
    def psigma(self, psigma):
        self.ptau = psigma*self.beta0

    tau = property(lambda self: self.zeta/self.beta)

    @tau.setter
    def tau(self, tau):
        self.zeta = self.beta*tau

    sigma = property(lambda self: (self.beta0/self.beta)*self.zeta)

    @sigma.setter
    def sigma(self, sigma):
        self.zeta = self.beta/self.beta0*sigma

    @property
    def ptau(self):
        sqrt = self._m.sqrt
        return sqrt(self.delta**2+2*self.delta + 1/self.beta0**2)-1/self.beta0

    @ptau.setter
    def ptau(self, ptau):
        sqrt = self._m.sqrt
        self.delta = sqrt(ptau**2+2*ptau/self.beta0+1)-1

    mass0 = property(lambda self: self._mass0)

    @mass0.setter
    def mass0(self, mass0):
        new = self._f1(mass0, self.p0c)

    beta0 = property(lambda self: self._beta0)

    @beta0.setter
    def beta0(self, beta0):
        new = self._f3(self.mass0, beta0)
        self._update_ref(*new)
        self._update_particles(*new)

    gamma0 = property(lambda self: self._gamma0)

    @gamma0.setter
    def gamma0(self, gamma0):
        new = self._f4(self.mass0, gamma0)
        self._update_ref(*new)
        self._update_particles(*new)

    p0c = property(lambda self: self._p0c)

    @p0c.setter
    def p0c(self, p0c):
        new = self._f1(self.mass0, p0c)
        self._update_ref(*new)
        self._update_particles(*new)

    energy0 = property(lambda self: self._energy0)

    @energy0.setter
    def energy0(self, energy0):
        new = self._f2(self.mass0, energy0)
        self._update_ref(*new)
        self._update_particles(*new)
    mratio = property(lambda self: self._mratio)

    @mratio.setter
    def mratio(self, mratio):
        Px, Py, Energy, Pc = self.Px, self.Py, self.Energy, self.Pc
        self._ptau = Energy/(mratio*energy0)-1
        self._delta = Energy/(delta*energy0)-1
        self.px = Px/(p0c*mratio)
        self.py = Py/(p0c*mratio)
        self._chi = self._qratio/mratio
        self._mratio = mratio
    qratio = property(lambda self: self._qratio)

    @qratio.setter
    def qratio(self, qratio):
        self._chi = qratio/self._mratio
        self._qratio = qratio
    chi = property(lambda self: self._chi)

    @chi.setter
    def chi(self, chi):
        self._qratio = self._chi*self._mratio
        self._chi = chi

    def _update_ref(self, mass0, beta0, gamma0, p0c, energy0):
        self._mass0 = mass0
        self._beta0 = beta0
        self._gamma0 = gamma0
        self._p0c = p0c
        self._energy0 = energy0

    def _update_particles(self, mass0, beta0, gamma0, p0c, energy0):
        if self._update_coordinates:
            Px = self.Px
            Py = self.Py
            Energy = self.Energy
            Pc = self.Pc
            mratio = mass/mass0
            self._mratio = mratio
            self._chi = self._qratio/mratio
            self._ptau = Energy/(mratio*energy0)-1
            self._delta = Energy/(delta*energy0)-1
            self.px = Px/(p0c*mratio)
            self.py = Py/(p0c*mratio)

    def __repr__(self):
        out = f"""\
        mass0   = {self.mass0}
        p0c     = {self.p0c}
        energy0 = {self.energy0}
        beta0   = {self.beta0}
        gamma0  = {self.gamma0}
        s       = {self.s}
        x       = {self.x}
        px      = {self.px}
        y       = {self.y}
        py      = {self.py}
        zeta    = {self.zeta}
        delta   = {self.delta}
        ptau    = {self.ptau}
        mratio  = {self.mratio}
        qratio  = {self.qratio}
        chi     = {self.chi}"""
        return out



import numpy as np
import math
import numba

from scipy.constants import e as qe
from scipy.constants import c as clight

from .gaussian_fields import get_Ex_Ey_Gx_Gy_gauss

from . import BB6D
from . import BB6Ddata

_factorial = np.array([1,
                       1,
                       2,
                       6,
                       24,
                       120,
                       720,
                       5040,
                       40320,
                       362880,
                       3628800,
                       39916800,
                       479001600,
                       6227020800,
                       87178291200,
                       1307674368000,
                       20922789888000,
                       355687428096000,
                       6402373705728000,
                       121645100408832000,
                       2432902008176640000])


class Element(object):
    def __init__(self, **nargs):
        not_allowed = set(nargs.keys())-set(self.__slots__)
        if len(not_allowed) > 0:
            cname = self.__class__.__name__
            raise NameError(f"{not_allowed} not allowed by {cname}")
        for name, default in zip(self.__slots__, self.__defaults__):
            setattr(self, name, nargs.get(name, default))

    def _slots(self):
        names = self.__slots__
        return zip(names, [getattr(self, name) for name in names])

    def __repr__(self):
        args = [f"{name}={val}" for name, val in self._slots()]
        args = ", ".join(args)
        return f"{self.__class__.__name__}({args})"

    def _asdict(self):
        return dict(self._slots())


class Drift(Element):
    __slots__ = ('length',)
    __units__ = ('meter',)
    __defaults__ = (0,)

    def track(self, p):
        length = self.length
        rpp = p.rpp
        xp = p.px*rpp
        yp = p.py*rpp
        p.x += xp*length
        p.y += yp*length
        p.zeta += length*(p.rvv-(1+(xp**2+yp**2)/2))
        p.s += length


class DriftExact(Element):
    __slots__ = ('length',)
    __units__ = ('meter',)
    __defaults__ = (0,)

    def track(self, p):
        sqrt = p._m.sqrt
        length = self.length
        opd = 1 + p.delta
        lpzi = length / sqrt(opd**2 - p.px**2 - p.py**2)
        p.x += p.px*lpzi
        p.y += p.py*lpzi
        p.zeta += p.rvv*length - opd*lpzi
        p.s += length


class Multipole(Element):
    __slots__ = ('knl', 'ksl', 'hxl', 'hyl', 'length')
    __units__ = ('meter^-order', 'meter^-order', 'radian', 'radian', 'meter')
    __defaults__ = ([], [], 0, 0, 0)

    @property
    def order(self):
        return max(len(self.knl), len(self.ksl))-1

    def track(self, p):
        order = self.order
        length = self.length
        knl = np.array(self.knl)
        ksl = np.array(self.ksl)
        if len(knl) < len(ksl):
            nknl = np.zeros(order+1, dtype=knl.dtype)
            nknl[:len(knl)] = knl
            knl = nknl
        elif len(knl) > len(ksl):
            nksl = np.zeros(order+1, dtype=ksl.dtype)
            nksl[:len(ksl)] = ksl
            ksl = nksl
        x = p.x
        y = p.y
        chi = p.chi
        dpx = knl[order]
        dpy = ksl[order]
        for ii in range(order, 0, -1):
            zre = (dpx*x-dpy*y)/ii
            zim = (dpx*y+dpy*x)/ii
            dpx = knl[ii-1]+zre
            dpy = ksl[ii-1]+zim
        dpx = -chi*dpx
        dpy = chi*dpy
        # curvature effect kick
        hxl = self.hxl
        hyl = self.hyl
        delta = p.delta
        if (hxl != 0 or hyl != 0):
            b1l = chi*knl[0]
            a1l = chi*ksl[0]
            hxlx = hxl*x
            hyly = hyl*y
            if (length > 0):
                hxx = hxlx/length
                hyy = hyly/length
            else:  # non physical weak focusing disabled (SixTrack mode)
                hxx = 0
                hyy = 0
            dpx += hxl + hxl*delta - b1l*hxx
            dpy -= hyl + hyl*delta - a1l*hyy
            p.zeta -= chi*(hxlx-hyly)
        p.px += dpx
        p.py += dpy


class XYShift(Element):
    """shift of the reference"""
    __slots__ = ('dx', 'dy')
    __units__ = ('meter', 'meter')
    __defaults__ = (0, 0)

    def track(self, p):
        p.x -= self.dx
        p.y -= self.dy


class SRotation(Element):
    """anti-clockwise rotation of the reference frame"""
    __slots__ = ('angle',)
    __units__ = ('degree',)
    __defaults__ = (0,)

    def track(self, p):
        deg2rag = p._m.pi/180
        cz = p._m.cos(self.angle*deg2rag)
        sz = p._m.sin(self.angle*deg2rag)
        xn = cz*p.x + sz*p.y
        yn = -sz*p.x + cz*p.y
        p.x = xn
        p.y = yn
        xn = cz*p.px + sz*p.py
        yn = -sz*p.px + cz*p.py
        p.px = xn
        p.py = yn


class Cavity(Element):
    __slots__ = ('voltage', 'frequency', 'lag')
    __units__ = ('volt', 'hertz', 'degree')
    __defaults__ = (0, 0, 0)

    def track(self, p):
        sin = p._m.sin
        pi = p._m.pi
        k = 2*pi*self.frequency/p.clight
        tau = p.zeta/p.rvv/p.beta0
        phase = self.lag*pi/180-k*tau
        p.add_to_energy(p.chi*self.voltage*sin(phase))


class RFMultipole(Element):
    __slots__ = ('voltage', 'frequency', 'knl', 'ksl', 'pn', 'ps')
    __units__ = ('volt', 'hertz', [], [], [], [])
    __defaults__ = (0, 0, 0, 0, 0)


class Line(Element):
    __slots__ = ('elements',)
    __defaults__ = ([],)

    def track(self, p):
        for el in self.elements:
            el.track(p)

    def track_elem_by_elem(self, p):
        out = []
        for el in self.elements:
            out.append(p.copy())
            el.track(p)
        return out


class Monitor(Element):
    __slots__ = ('data',)
    __defaults__ = ([],)

    def track(self, p):
        self.data.append(p.copy)


class BeamBeam4D(Element):
    __slots__ = ('q_part', 'N_part', 'sigma_x', 'sigma_y', 'beta_s',
                 'min_sigma_diff', 'Delta_x', 'Delta_y', 'Dpx_sub', 'Dpy_sub', 'enabled')
    __units__ = ('C', [], 'm', 'm', [],
                 'm', 'm', 'm', [], [], [])
    __defaults__ = (0., 0., 0., 0., 0.,
                    0., 0., 0., 0., 0., True)

    def track(self, p):
        if self.enabled:
            charge = p.qratio*p.q0*qe
            x = p.x - self.Delta_x
            px = p.px
            y = p.y - self.Delta_y
            py = p.py

            chi = p.chi

            beta = p.beta0/p.rvv
            p0c = p.p0c*qe

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(x, y, self.sigma_x, self.sigma_y,
                                           min_sigma_diff=1e-10, skip_Gs=True, mathlib=p._m)

            fact_kick = chi * self.N_part * self.q_part * charge * \
                (1. + beta * self.beta_s)/(p0c*(beta + self.beta_s))

            px += (fact_kick*Ex - self.Dpx_sub)
            py += (fact_kick*Ey - self.Dpy_sub)

            p.px = px
            p.py = py

    def tobuffer(self):

        buffer_list = []

        buffer_list.append(np.array([self.q_part], dtype=np.float64))
        buffer_list.append(np.array([self.N_part], dtype=np.float64))
        buffer_list.append(np.array([self.sigma_x], dtype=np.float64))
        buffer_list.append(np.array([self.sigma_y], dtype=np.float64))
        buffer_list.append(np.array([self.beta_s], dtype=np.float64))
        buffer_list.append(np.array([self.min_sigma_diff], dtype=np.float64))
        buffer_list.append(np.array([self.Delta_x], dtype=np.float64))
        buffer_list.append(np.array([self.Delta_y], dtype=np.float64))
        buffer_list.append(np.array([self.Dpx_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dpy_sub], dtype=np.float64))
        buffer_list.append(BB6Ddata.int_to_float64arr(
            {True: 1, False: 0}[self.enabled]))

        buf = np.concatenate(buffer_list)

        return buf


class BeamBeam6D(Element):
    __slots__ = (['q_part', 'N_part_tot', 'sigmaz', 'N_slices', 'min_sigma_diff', 'threshold_singular',
                  'phi', 'alpha',
                  'Sig_11_0', 'Sig_12_0', 'Sig_13_0',
                  'Sig_14_0', 'Sig_22_0', 'Sig_23_0',
                  'Sig_24_0', 'Sig_33_0', 'Sig_34_0', 'Sig_44_0',
                  'delta_x', 'delta_y',
                  'x_CO', 'px_CO', 'y_CO', 'py_CO', 'sigma_CO', 'delta_CO',
                  'Dx_sub', 'Dpx_sub', 'Dy_sub', 'Dpy_sub', 'Dsigma_sub', 'Ddelta_sub',
                  'enabled'])
    __units__ = tuple(len(__slots__)*[[]])
    __defaults__ = tuple(len(__slots__)*[0.])

    def track(self, p):
        if self.enabled:
            bb6data = BB6Ddata.BB6D_init(
                self.q_part, self.N_part_tot, self.sigmaz, self.N_slices, self.min_sigma_diff, self.threshold_singular,
                self.phi, self.alpha,
                self.Sig_11_0, self.Sig_12_0, self.Sig_13_0,
                self.Sig_14_0, self.Sig_22_0, self.Sig_23_0,
                self.Sig_24_0, self.Sig_33_0, self.Sig_34_0, self.Sig_44_0,
                self.delta_x, self.delta_y,
                self.x_CO, self.px_CO, self.y_CO, self.py_CO, self.sigma_CO, self.delta_CO,
                self.Dx_sub, self.Dpx_sub, self.Dy_sub, self.Dpy_sub, self.Dsigma_sub, self.Ddelta_sub,
                self.enabled)
            x_ret, px_ret, y_ret, py_ret, zeta_ret, delta_ret = BB6D.BB6D_track(
                p.x, p.px, p.y, p.py, p.zeta, p.delta, p.q0*qe, p.p0c/clight*qe, bb6data)
            self._last_bb6data = bb6data
            p.x = x_ret
            p.px = px_ret
            p.y = y_ret
            p.py = py_ret
            p.zeta = zeta_ret
            p.delta = delta_ret


classes = [cls for cls in globals().values() if isinstance(cls, type)]
elements = [cls for cls in classes if issubclass(cls, Element)]
__all__ = [cls.__name__ for cls in elements]
__all__.append('element_types')

element_types = dict((cls.__name__, cls) for cls in elements)



import numpy as np

# I program as close as possible to C...


def boost(x, px, y, py, sigma, delta, parboost):

    sphi = parboost.sphi
    cphi = parboost.cphi
    tphi = parboost.tphi
    salpha = parboost.salpha
    calpha = parboost.calpha

    h = delta + 1. - np.sqrt((1.+delta)*(1.+delta)-px*px-py*py)

    px_st = px/cphi-h*calpha*tphi/cphi
    py_st = py/cphi-h*salpha*tphi/cphi
    delta_st = delta - px*calpha*tphi-py*salpha*tphi+h*tphi*tphi

    pz_st = np.sqrt((1.+delta_st)*(1.+delta_st)-px_st*px_st-py_st*py_st)
    hx_st = px_st/pz_st
    hy_st = py_st/pz_st
    hsigma_st = 1.-(delta_st+1)/pz_st

    L11 = 1.+hx_st*calpha*sphi
    L12 = hx_st*salpha*sphi
    L13 = calpha*tphi

    L21 = hy_st*calpha*sphi
    L22 = 1.+hy_st*salpha*sphi
    L23 = salpha*tphi

    L31 = hsigma_st*calpha*sphi
    L32 = hsigma_st*salpha*sphi
    L33 = 1./cphi

    x_st = L11*x + L12*y + L13*sigma
    y_st = L21*x + L22*y + L23*sigma
    sigma_st = L31*x + L32*y + L33*sigma

    return x_st, px_st, y_st, py_st, sigma_st, delta_st


def inv_boost(x_st, px_st, y_st, py_st, sigma_st, delta_st, parboost):

    sphi = parboost.sphi
    cphi = parboost.cphi
    tphi = parboost.tphi
    salpha = parboost.salpha
    calpha = parboost.calpha

    pz_st = np.sqrt((1.+delta_st)*(1.+delta_st)-px_st*px_st-py_st*py_st)
    hx_st = px_st/pz_st
    hy_st = py_st/pz_st
    hsigma_st = 1.-(delta_st+1)/pz_st

    Det_L = 1./cphi + (hx_st*calpha + hy_st*salpha-hsigma_st*sphi)*tphi

    Linv_11 = (1./cphi + salpha*tphi*(hy_st-hsigma_st*salpha*sphi))/Det_L
    Linv_12 = (salpha*tphi*(hsigma_st*calpha*sphi-hx_st))/Det_L
    Linv_13 = -tphi*(calpha - hx_st*salpha*salpha*sphi +
                     hy_st*calpha*salpha*sphi)/Det_L

    Linv_21 = (calpha*tphi*(-hy_st + hsigma_st*salpha*sphi))/Det_L
    Linv_22 = (1./cphi + calpha*tphi*(hx_st-hsigma_st*calpha*sphi))/Det_L
    Linv_23 = -tphi*(salpha - hy_st*calpha*calpha*sphi +
                     hx_st*calpha*salpha*sphi)/Det_L

    Linv_31 = -hsigma_st*calpha*sphi/Det_L
    Linv_32 = -hsigma_st*salpha*sphi/Det_L
    Linv_33 = (1. + hx_st*calpha*sphi + hy_st*salpha*sphi)/Det_L

    x_i = Linv_11*x_st + Linv_12*y_st + Linv_13*sigma_st
    y_i = Linv_21*x_st + Linv_22*y_st + Linv_23*sigma_st
    sigma_i = Linv_31*x_st + Linv_32*y_st + Linv_33*sigma_st

    h = (delta_st+1.-pz_st)*cphi*cphi

    px_i = px_st*cphi+h*calpha*tphi
    py_i = py_st*cphi+h*salpha*tphi

    delta_i = delta_st + px_i*calpha*tphi + py_i*salpha*tphi - h*tphi*tphi

    return x_i, px_i, y_i, py_i, sigma_i, delta_i



import setuptools

setuptools.setup(
        name='pysixtrack',
        version='0.0.0',
        description='6D Tracking Code',
        author='Riccardo De Maria',
        author_email='riccardo.de.maria@cern.ch',
        url='https://github.com/rdemaria/pysixtrack',
        packages=['pysixtrack'],
        package_dir={'pysixtrack': 'pysixtrack'},
        install_requires=['numpy']
)
